/*
    semu - Sega Saturn Emulator
    Copyright (C) 1998  Chuck Mason

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Chuck Mason IV			chuckjr@sundial.net
    477 Moffat Loop
    Oviedo, Fl 32765
*/

#include <string.h>
#include <mntent.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <malloc.h>

#ifndef DOS
#	include <sys/ioctl.h>
#endif

#include <signal.h>
#include <fcntl.h>

#if !defined(DOS) && !defined(NOCDROM)
#	include <linux/cdrom.h>
#	include <linux/errno.h>
#endif

#include "main.h"
#include "header.h"
#include "game.h"
#include "bios.h"
#include "error.h"

#include "cpu/sh2/sh2proc.h"
#include "cpu/sh2/opcodes.h"
#include "cpu/sh2/memory.h"

int saturn_cdrom;
FILE *saturn_cdrom_file;

char *cddevice;
char *romfile;
unsigned long romfile_len;

int emulate_bios(char *cddev)
{
	if(use_cdrom == 1) {
#if !defined(DOS) && !defined(NOCDROM)
		unsigned long i;
		
		if((saturn_cdrom_file = fopen(cddev, "r")) == NULL) 
 			return error(ERR_NOOPEN, cddev);
		saturn_cdrom = fileno(saturn_cdrom_file);
		
		if(read_header(saturn_cdrom) != RET_SUCCESS) 
			return RET_FAILURE;
	
		if(validify_header() != RET_SUCCESS)
       			return RET_FAILURE;
        
		if(show_cdrom_info) {
			display_header();
			return RET_SUCCESS;
		} 
	
		if(confirm_security() == RET_SUCCESS)
			printf("Security Code Test Passed.\n");
		else
			return RET_FAILURE;

		if(confirm_unitedstates_security() == RET_SUCCESS)
			printf("United States Area Code Test Passed.\n");
		else
			return RET_FAILURE;
		
		if(play_music)
			return cdrom_play_track(play_music);
			
		printf("\nPlaying game \"%s\".\nCPU Is partly emulated but cdrom is not playable, yet.\n", 
			fmttitle(current_game.header.title));			

		if(read_aip(saturn_cdrom) != RET_SUCCESS)
			return RET_FAILURE;
		
		printf("Reading nto SH2 Memory... security, ");
		fflush(stdout);
		
		for(i = 0; i < SECURITY_CODE_LEN; i++)
			sh2_write_byte(i, current_game.security_code[i]);
		printf("area code, ");
		fflush(stdout);
		for(i = 0; i < (AREA_CODE_LEN * 8); i++)
			sh2_write_byte(i+SECURITY_CODE_LEN, current_game.area_codes[i]);
		printf("aip... ");
		fflush(stdout);
		for(i = 0; i < APPLICATION_INITIAL_PROGRAM_SIZE; i++)
			sh2_write_byte(SECURITY_CODE_LEN+(AREA_CODE_LEN*8)+i, current_game.application_initial_program[i]);
			
		printf("done.\n");
			
#elif defined(NOCDROM) && !defined(DOS)
		printf("Support for CDROM was disabled in compile.\n");
		return RET_SUCCESS;
#elif defined(DOS)
		printf("No support for CDROM in DOS! Hah, hah.\n");
		return RET_FAILURE;
#endif                
	} else if(use_romfile == 1) {
		FILE *fp;
		unsigned long len, i;

		if(play_music) {
			printf("Sorry, music can only be played from cd!\n");
			return RET_FAILURE;
		}
		
		if(show_cdrom_info) {
			printf("Sorry, cannot show cdrom info if cdrom is not used!\n");
			return RET_FAILURE;
		}
					
		if((fp = fopen(romfile, "rb")) == NULL) {
			printf("Cannot open rom file \"%s\"\n", romfile);
			free(romfile);
			return RET_FAILURE;
		}
		
		printf("File: [%s]\n", romfile);
		
		fseek(fp, 0, SEEK_END);
		romfile_len = len = ftell(fp);
		rewind(fp);
		
		for(i = 1; i <= len; i++) {
			unsigned char c;
			c = fgetc(fp);
			sh2_write_byte(i - 1, c);
		}
		
		fclose(fp);
	}

	sh2_build_instruction_list();
	sh2_reset_cpu();

	if(use_cdrom)
		sh2.R[15] = current_game.header.stack_m;

	if(sh2.R[15] == 0)
		sh2.R[15] = 0x06001FFF;

	printf("\nStack Pointer at %08X\n", sh2.R[15]);
	
#ifdef DEBUG
	if(use_sh2_debugger) {
		if(use_cdrom)
			sh2_debugger(APPLICATION_INITIAL_PROGRAM_SIZE+(AREA_CODE_LEN*8)+SECURITY_CODE_LEN);
		else if(use_romfile)
			sh2_debugger(romfile_len);
		else
			sh2_debugger(0);
	} 
#endif
	if(!use_sh2_debugger) {
		while(sh2_processor_sleep == 0 && sh2_invalid_opcode_occ != 1 && ((sh2.PC < romfile_len) || (use_romfile == 0))) {
#ifdef DEBUG
			printf("%08X %04X: ", sh2.PC, sh2_read_word(sh2.PC));
#endif
			sh2_execute_instruction(sh2_read_word(sh2.PC));
		}
	}
	
	return RET_SUCCESS;
}

int confirm_security()
{
	/* This just basically returns 1 if current_game.security_code
	   matches sega_security_code exactly */
	int i;
	for(i = 0; i < SECURITY_CODE_LEN; i++) {
		if(current_game.security_code[i] != sega_security_code[i])
			return error(ERR_SECURITY_CODE_NOMATCH);
	}
	return RET_SUCCESS;
}

int confirm_unitedstates_security()
{
	int i, j;
	
	for(i = 0; i < 8; i++) {
		int allmatch = 1;
		for(j = 0; j < AREA_CODE_LEN; j++) {
			if(current_game.area_codes[i*j] != area_code_us[j]) {
				allmatch = 0;
				break;
			}
		}
		if(allmatch == 1) {
			/* We have a winner!! */
			return RET_SUCCESS;
		}
	}
	return error(ERR_AREA_CODE_NOMATCH);
}

#if !defined(DOS) && !defined(NOCDROM)
int cdrom_play_track(int start)
{
	struct cdrom_ti ti;
	ti.cdti_trk0 = start;
	ti.cdti_ind0 = 1;
	ti.cdti_trk1 = start+10;
	ti.cdti_ind0 = 1;
	
	if(ioctl(saturn_cdrom, CDROMPLAYTRKIND, &ti) == -1) 
		return error(ERR_NOPLAY_MUSIC);
	else {
		char c = '\\';
		printf("Press Control-C to break play.\n");
		while(1) {
			printf("%c\r", c); 
			fflush(stdout);
			switch(c) {
				case '\\':
					c = '|';
					break;
				case '|':
					c = '/';
					break;
				case '/':
					c = '-';
					break;
				case '-':
					c = '\\';
					break;
			}
			sleep(1);
		}
	}
	
	return RET_SUCCESS;
}
#endif
