/*
    semu - Sega Saturn Emulator
    Copyright (C) 1998  Chuck Mason

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Chuck Mason IV                      chuckjr@sundial.net
    477 Moffat Loop
    Oviedo, Fl 32765
*/

#ifdef DEBUG

#include <stdio.h>
#include <stdlib.h>

#include "sh2proc.h"
#include "opcodes.h"
#include "memory.h"
#include "../../gpl.h"

void sh2_debug_exit();
void sh2_debug_step();
void sh2_debug_go();
void sh2_debug_next();
void sh2_debug_regs();
void sh2_debug_help();
void sh2_debug_reset();
void sh2_debug_version();
void sh2_debug_skip();
void sh2_debug_dumpmem();
void sh2_debug_writemem();
void sh2_debug_swapb();
void sh2_debug_roml();
void sh2_debug_execute();

static int sh2_debug_status;
static int rom_len;
static char inputline[512];

int sh2_in_debugger = 0;

typedef struct _cmds {
	char cmd;
	char *cmdinfo;
	void (*func)();
} command;

command commands[] = {
	{ 'e', "Exit the debugger", sh2_debug_exit },
	{ 'q', "Exit the debugger (also e)", sh2_debug_exit },
	{ 's', "Step one instruction", sh2_debug_step },
	{ 'k', "Skip current instruction", sh2_debug_skip },
	{ 'g', "Continue execution until breakpoint or invalid opcode", sh2_debug_go },
	{ 'o', "Display next opcode", sh2_debug_next },
	{ 'p', "Display SH2 Processor Registers", sh2_debug_regs },
	{ 'd', "Dump memory. Can also use \"d x\" where [x] is a hex value", sh2_debug_dumpmem },
	{ 'w', "Write memory. \"w x a\" Writes hex value A at hex address X", sh2_debug_writemem },
	{ 'r', "Reset the SH2 processor", sh2_debug_reset },
	{ 'h', "Displays this help", sh2_debug_help },
	{ 'b', "Swap every other byte (reverse endian)", sh2_debug_swapb },
	{ '?', "Displays this help (also h)", sh2_debug_help },
	{ 'v', "Displays debugger version", sh2_debug_version },
	{ 'i', "Execute instruction X \"i X\"", sh2_debug_execute },
	{ 0x00, (char *)NULL, sh2_debug_exit }
};


int sh2_debugger(unsigned long romlen)
{
	char in[512];
	char cmd, oldcmd = 0;
	int i;

	sh2_in_debugger = 1;	
	sh2_debug_status = 0;
	rom_len = romlen;

	printf("Entering SH2 debugger...\n");
	while(sh2_debug_status == 0) {
		printf("[1;34mSH[1;31m2[1;37m>[0;37m ");
		fflush(stdout);
		
		gets(in);
		
		cmd = in[0];
		memset(inputline, 0, 511);
		strcpy(inputline, &in[1]);
		
		fflush(stdin);

		if(strncasecmp(in, "show copying", 12) == 0) {
			show_copying();
			continue;
		} else if(strncasecmp(in, "show warranty", 13) == 0) {
			show_warranty();
			continue;
		} else if(strncasecmp(in, "set r", 5) == 0) {
			unsigned long a;
			short int d;
			sscanf(in, "set r%d %08X\n", &d, &a);
			sh2.R[d] = a;
			continue;
		}
		
		if(cmd == 0 && oldcmd != 0)
			cmd = oldcmd;
		else if(cmd == 0 && oldcmd == 0)
			continue;
		else if(cmd == ' ')
			continue;
		for(i = 0; commands[i].cmd != 0x00; i++) {
			if(cmd == commands[i].cmd) {
				commands[i].func();
				break;
			}
		}
		
		if(commands[i].cmd == 0x00) 
			printf("%c - invalid command\n", cmd);
		
		oldcmd = cmd;
	}

	sh2_in_debugger = 0;	
	return sh2_debug_status;
}

void sh2_debug_exit()
{
	sh2_debug_status = 1;
}

void sh2_debug_step()
{
	if(sh2.PC >= rom_len && (rom_len != 0)) {
		printf("PC has surpassed the length of the rom image.\n");
		return;
	}
	                                                
	printf("%08X %04X: ", sh2.PC, sh2_read_word(sh2.PC));
	sh2_execute_instruction(sh2_read_word(sh2.PC));
}

void sh2_debug_go() 
{
	sh2_invalid_opcode_occ = 0;

	if((sh2.PC >= rom_len) && (rom_len != 0)) {
		printf("PC has surpassed the length of the rom image.\n");
		return;
	}

	while(sh2_invalid_opcode_occ == 0 && ((rom_len == 0) || (sh2.PC < rom_len))) {
		printf("%08X %04X: ", sh2.PC, sh2_read_word(sh2.PC));
		sh2_execute_instruction(sh2_read_word(sh2.PC));
	}
	sh2_invalid_opcode_occ = 0;
}

void sh2_debug_next()
{
	char opstr[512];
	sh2_format_instruction_string(sh2_read_word(sh2.PC), opstr);
	printf("%08X %04X: %s\n", sh2.PC, sh2_read_word(sh2.PC), opstr);
}

void sh2_debug_regs()
{
	printf("R00..: %08X R01..: %08X R02..: %08X R03..: %08X\n",
		sh2.R[0], sh2.R[1], sh2.R[2], sh2.R[3]);
	printf("R04..: %08X R05..: %08X R06..: %08X R07..: %08X\n",
		sh2.R[4], sh2.R[5], sh2.R[6], sh2.R[7]);
	printf("R08..: %08X R09..: %08X R10..: %08X R11..: %08X\n",
		sh2.R[8], sh2.R[9], sh2.R[10], sh2.R[11]);
	printf("R12..: %08X R13..: %08X R14..: %08X R15..: %08X\n",
		sh2.R[12], sh2.R[13], sh2.R[14], sh2.R[15]);
	printf("MACH.: %08X MACL.: %08X GBR..: %08X VBR..: %08X\n",
		sh2.MACH, sh2.MACL, sh2.GBR, sh2.VBR);
	printf("PC...: %08X PR...: %08X SR...: %08X FLAG.: %c%c%c%c\n",
		sh2.PC, sh2.PR, sh2.SR, (M == 1) ? 'M' : 'm',
			    (Q == 1) ? 'Q' : 'q',
			    (S == 1) ? 'S' : 's',
			    (T == 1) ? 'T' : 't');
	printf("@PC..: %04X\n", sh2_read_word(sh2.PC));
}

void sh2_debug_help()
{
	int i;
	for(i = 0; commands[i].cmd != 0x00; i++)
		printf("%c - %s\n", commands[i].cmd, commands[i].cmdinfo);
	printf("show warranty - Various kinds of warranty you do not have\n");
	printf("show copying - Conditions for redistributing copies of semu\n");
}

void sh2_debug_reset()
{
	sh2_reset_cpu();
}

void sh2_debug_version()
{
	printf("SH2 Debugger v0.1.3 Copyright (C) 1998 Chuck Mason\n");
	printf("For terms of warranty or copying, type \"show warranty\"\n");
	printf("or \"show copying\". All other information, read GPL.\n");
}

void sh2_debug_skip()
{
	printf("Opcode $%04X Skipped\n", sh2_read_word(sh2.PC));
	sh2.PC += 2;
	printf("PC = %08X\n", sh2.PC);
}

void putschar(unsigned char c)
{
	if((c < 32) || (c > 126)) 
		putchar('.');
        else 
        	putchar(c);
}
        
void sh2_debug_dumpmem()
{
	int i, cnt;
	unsigned long where;
	static unsigned long lastaddr = 0;
	where = lastaddr;

	sscanf(inputline, " %X", &where);

	for(i = 0; i < 5; i++) {
		printf("%08X: ", where);
		for(cnt = 0; cnt < 16; cnt++) {
			printf("%02X", sh2_read_byte(where + cnt));
			if(cnt == 7)
				putchar('-');
			else
				putchar(' ');
		}
		printf("  ");
		for(cnt = 0; cnt < 16; cnt++) 
			putschar(sh2_read_byte(where + cnt));
		printf("\n");
		where += 0x10;
	}
	lastaddr = where;
}
		
void sh2_debug_writemem()
{
	int i;
	unsigned long where, what;
	where = 0;
	what = 0;
	
	if(inputline[0] == '\0') {
		printf("Usage: w <hex address> <byte to store>\n");
		return;
	}
	
	sscanf(inputline, "%X %X", &where, &what);
	what &= 0xFF;
	
	sh2_write_byte(where, what);
}

void sh2_debug_execute()
{
	unsigned long op;
	char opstr[512];
	
	if(inputline[0] == '\0') {
		printf("Usage: i <opcode as 2 byte hex>\n");
		return;
	}
	
	sscanf(inputline, "%X", &op);
	
	printf("Executing %04X: ", op);
	sh2_execute_instruction(op);
	
	/* Lame, because SLEEP doesn't increment PC */
	if(strcmp(sh2_instructions[op].instrformat, "SLEEP") != 0)
		sh2.PC -= 2;
		
	/* Clear sh2_invalid_opcode_occ [just in case] */
	sh2_invalid_opcode_occ = 0;
}

void sh2_debug_swapb()
{
	if(rom_len == 0) {
		printf("Unknown length of ROM image, unable to swap. Enter debugger with -2.\n");
		return;
	} else {
		unsigned long i;
		
		printf("Swapping...");
		fflush(stdout);
		
		for(i = 0; i < rom_len; i += 2) {
			char a, b;
			a = sh2_read_byte(i);
			b = sh2_read_byte(i+1);
			sh2_write_byte(i, b);
			sh2_write_byte(i + 1, a);
		}
		printf("done.\n");
	}
}

void sh2_debug_roml()
{
	printf("The length of rom is %lu\n", rom_len);
}

#else

#warning "No Debugger Available"

int sh2_debugger(unsigned long i)
{
	printf("Debugger not available!\n");
	return 0;
}

#endif